/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
/*
 * Copyright 2004 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */


/*
 * This program compresses a file using Zlib. This is used to generate
 * compressed files that can be read via the lofi decompression feature.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <netinet/in.h>
#include <inttypes.h>
#include <locale.h>
#include <zlib.h>

/*
 * Flag values indicate whether a block is compressed or not
 */
#define COMPRESSED 1
#define UNCOMPRESSED 0

/*
 * Device block size
 */
#define DEV_BSIZE 512

/*
 * Size of the compressed segment header in bytes
 */
#define SEGHDR 1

/*
 * The signature that identifies our compressed file
 */
#define SIGNATURE "compzlib"

#define CACHE_SIZE 10485760
static unsigned char cache[CACHE_SIZE];


int
main(int argc, char *argv[]) 
{
	char *fname, *tfname1, *tfname2, *cmdline, *sign;
	uint32_t segsize, entries, count, csegsize;
	uint32_t *index, offset, poffset;
	unsigned long blksz;
	int fd, tfd, rv;
	struct stat sbuf;
	unsigned char *data_seg, *data_seg1, *compressed_seg;
	ssize_t rbytes, wbytes, lastread;
	size_t cmplen, mingap, off;
	uint32_t wint;
	unsigned char flag;

        (void) setlocale(LC_ALL, "");
#if !defined(TEXT_DOMAIN)       /* Should be defined by cc -D */
#define TEXT_DOMAIN "SYS_TEST"  /* Use this only if it weren't */
#endif
        (void) textdomain(TEXT_DOMAIN);

	if (argc < 4) {
		printf("Usage: lcompress <filename> <segment size> <threshold>\n\n");
		exit(1);
	}

	/*
	 * fname - filename to comopress
	 * segsize - the file will be split up into segments of this size
	 *           which will be individually compressed.
	 */
	sign = SIGNATURE;
	fname = argv[1];
	segsize = atol(argv[2]);

	/*
	 * Minimum threshold size in bytes. The compressed segment size must be
	 * at least mingap less than segsize 
	 */
	mingap = atol(argv[3]);

	/*
	 * Size of the buffer to hold the compressed data. Slightly
	 * larger than the compression segment size to allow for
	 * zlib requirements.
	 */
	csegsize = segsize + (segsize >> 6);
	compressed_seg = (unsigned char *)malloc(csegsize + SEGHDR);
	data_seg = (unsigned char *)malloc(csegsize);

	if (compressed_seg == NULL || data_seg == NULL) {
		perror("Unable to allocate memory ");
		exit(1);
	}

	fd = open(fname, O_RDONLY, 0);
	if (fd == -1) {
		fprintf(stderr, "Unable to open file %s", fname);
		perror(" ");
		exit(1);
	}

	if (fstat(fd, &sbuf) == -1) {
		fprintf(stderr, "Stat failed for file %s", fname);
		perror(" ");
		close(fd);
		exit(1);
	}

	if (segsize % DEV_BSIZE) {
		fprintf(stderr, "ERROR: given segment size %ld ", segsize);
		fprintf(stderr, "is not a multiple of minimum I/O block");
		fprintf(stderr, " size %ld\n", segsize, DEV_BSIZE);
		close(fd);
		exit(1);
	}

	/*
	 * File names to hold intermediate and final data
	 */
	tfname1 = malloc(strlen(fname) + 10);
	tfname2 = malloc(strlen(fname) + 10);
	if (tfname1 == NULL || tfname2 == NULL) {
		perror(" ");
		close(fd);
		exit(1);
	}
	strcpy(tfname1, fname);
	strcpy(tfname2, fname);
	strcat(tfname1, ".data");
	strcat(tfname2, ".zlib");

	tfd = open(tfname1, O_WRONLY | O_CREAT | O_SYNC | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (tfd == -1) {
		fprintf(stderr, "Unable to open file %s for writing\n", tfname1);
		perror(" ");
		close(fd);
		exit(1);
	}

	/*
	 * How many index entries will we need ?
	 */
	entries = sbuf.st_size / segsize + 2;
	index = malloc(sizeof(uint32_t) * entries);

	if (index == NULL) {
		perror(" ");
		close(fd);
		exit(1);
	}

	printf("Index will have %u entries\n", entries);
	offset = 0;
	poffset = 0;
	lastread = segsize;
	count = 0;

	/*
	 * Read in a segment of data from the given file
	 */
	while ((rbytes = read(fd, (void *)data_seg, segsize)) > 0) {

		if (lastread < segsize) {
			fprintf(stderr, "Segment size %u ", lastread);
			fprintf(stderr, "read at offset %u is less ", poffset);
			fprintf(stderr, "than given segsize %u\n", segsize);
			close(fd);
			close(tfd);
			exit(1);
		}

		blksz = csegsize-1;

		/*
		 * Compress the segment using Zlib
		 */
		rv = compress2(compressed_seg + SEGHDR, &blksz,
				data_seg, rbytes, 9);
		if (rv != Z_OK) {
			printf("Zlib error: code %d, count = %u\n", rv, count);
		}

		/*
		 * If length of compressed data is more than a threshold
		 * then store the segment in uncompressed form else store
		 * the compressed data. The flag is set accordingly.
		 *
		 * Storing this flag with the data reduces memory requirements
		 * for the index.
		 */
		cmplen = blksz;
		if (cmplen > segsize - mingap) {
			memcpy(compressed_seg + SEGHDR, data_seg, rbytes);
			flag = UNCOMPRESSED;
			cmplen = rbytes;
		} else {
			flag = COMPRESSED;
		}
		*compressed_seg = flag;
		wbytes = write(tfd, compressed_seg, cmplen+SEGHDR);
		index[count] = htonl(offset);

		poffset = offset;
		offset += wbytes;
		lastread = rbytes;
		count++;
	}

	close(fd);
	close(tfd);

	/*
	 * This is a sentinel index entry. It does not point to an
	 * actual compressed segment but helps in computing the size
	 * of the compressed segment. The size of each compressed
	 * segment is computed by subtracting the current index value
	 * from the next one - as the compressed blocks are stored
	 * sequentially.
	 */
	index[count++] = htonl(offset); /* sentinel */

	/*
	 * Open the actual compressed file.
	 */
	tfd = open(tfname2, O_WRONLY | O_CREAT | O_SYNC | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (tfd == -1) {
		perror("compress[8]: ");
		close(tfd);
		exit(1);
	}

	/*
	 * Write the header and the index
	 *
	 * signature
	 * segsize (user specified)
	 * number of index entries
	 */
	if (write(tfd, sign, 8) != 8) {
		perror("compress[8]: ");
		close(tfd);
		exit(1);
	}

	wint = htonl(segsize);
	if (write(tfd, &wint, 4) != sizeof(wint)) {
		perror("compress[10]: ");
		close(tfd);
		exit(1);
	}
	wint = htonl(count);
	if (write(tfd, &wint, 4) != sizeof(wint)) {
		perror("compress[10]: ");
		close(tfd);
		exit(1);
	}
	wint = htonl(lastread);
	if (write(tfd, &wint, 4) != sizeof(wint)) {
		perror("compress[11]: ");
		close(tfd);
		exit(1);
	}
	for (wint=0; wint < count; wint++) {
		if (write(tfd, index+wint, 4) != 4) {
			perror("compress[12]: ");
			close(tfd);
			exit(1);
		}
	}
	close(tfd);

	cmdline = (char *)malloc(strlen(tfname1) + strlen(tfname2) + 20);
	if (cmdline == NULL) {
		perror("compress[13]: ");
		exit(1);
	}

	/*
	 * Append the compressed data to our target compressed file.
	 */
	strcpy(cmdline, "/usr/bin/cat ");
	strcat(cmdline, tfname1);
	strcat(cmdline, " >> ");
	strcat(cmdline, tfname2);
	
	rv = system(cmdline);
	if (rv == -1) {
		perror("compress[14]: ");
		exit(1);
	} else if (rv != 0) {
		fprintf(stderr, "Failed to append %s to %s\n", tfname1, tfname2);
		exit(1);
	}

	/*
	 * Remove the intermediate compressed data file.
	 */
	strcpy(cmdline, "/usr/bin/rm ");
	strcat(cmdline, tfname1);
	system(cmdline);

	tfd = open(tfname2, O_RDONLY, 0);
	if (tfd == -1) {
		perror("compress[14]: ");
		exit(1);
	}
	if (fstat(tfd, &sbuf) == -1) {
		perror("compress[15]: ");
		close(tfd);
		exit(1);
	}
	close(tfd);

	/*
	 * Round up the compressed file size to be a multiple of DEV_BSIZE.
	 * lofi(7D) cribs otherwise.
	 */
	if ((off = sbuf.st_size % DEV_BSIZE) > 0) {
		int i;

		off = DEV_BSIZE - off;
		tfd = open(tfname2, O_WRONLY | O_APPEND, 0);
		if (tfd == -1) {
			perror("compress[16]: ");
			exit(1);
		}
		for (i=0; i<off; i++)
			data_seg[i] = '\0';
		if (write(tfd, data_seg, off) != off) {
			perror("compress[17]: ");
			close(tfd);
			exit(1);
		}
		close(tfd);
	}

	return (0);
}
